#ifndef COMPONENTS_RTSHADERSYSTEM_INCLUDE_OGRERTSHADERCONFIG_H_
#define COMPONENTS_RTSHADERSYSTEM_INCLUDE_OGRERTSHADERCONFIG_H_

#define RTSHADER_SYSTEM_BUILD_CORE_SHADERS

#define RTSHADER_SYSTEM_BUILD_EXT_SHADERS

#endif
